/**
 ** rtslamMainTaskCodels.cc
 **
 ** Codels called by execution task rtslamMainTask
 **
 ** Author: 
 ** Date: 
 **
 **/




/*

donc pour le module rtslam...
on garde init, start, stop
on a un poster de status en plus, qui contient un flag j'ai divergé, et un flag pret à bouger avec en plus un compte à rebours
on a aussi une tache qui fait un historique de l'odométrie (un poster pom générique)
si on diverge, on arrete le slam et on le signale dans le poster, en enregistrant la dernière bonne position avec sa date.
Eventuellement on continue à exporter avec l'odométrie.
Le superviseur voit ça et arrete le robot, puis refait un start de rtslam, qui se réinitialise avec l'odométrie après la dernière bonne position,
et repart, en indiquant dans le poster de status quand il est pret à tourner.

*/

#include <portLib.h>
#include "posterLib.h"

#include "server/rtslamHeader.h"
#include "jmath/jblas.hpp"

#include "rtslam/main.hpp"


// threshold in meters above which we stop because we consider we are lost
const int LARGE_UNCERTAINTY_THRESHOLD = 100;

static POSTER_ID pomFramePosterId = NULL;
static const char* pomPoster = "pomrtslamFramePos";


STATUS
rtslamMainTaskInit(int *report)
{
	// pomPos
	SDI_F->pomPos.kind = POM_ME_ABSOLUTE;
	SDI_F->pomPos.confidence = 0.5;
	SDI_F->pomPos.date2 = 0;
	memset(&(SDI_F->pomPos.main.var),0,sizeof(POM_EULER_VARIANCES));

	// status
	SDI_F->status.ready = RTSLAM_FALSE;
	SDI_F->status.error_code = eNoError;
	SDI_F->status.time_to_ready = 0.0; // TODO

	// other
	SDI_F->exportDelay = 0.0;

	return OK;
}


STATUS
rtslamMainTaskMain(int *report)
{
	if (!SDI_F->status.ready && SDI_F->status.error_code != eLargeUncertainty) return OK;

	// read pom date
	if (pomFramePosterId)
	{
		POM_SENSOR_POS framePos;
		if (posterRead(pomFramePosterId, 0, &framePos, sizeof(POM_SENSOR_POS)) == ERROR) {
			printf("[rtslam-genom] ERROR: cannot read pom tagging poster %s\n", pomPoster);
			*report = S_rtslam_CANNOT_READ_POM;
			return ERROR;
		}
		SDI_F->pomPos.date1 = framePos.date;
	} else
	{
		SDI_F->pomPos.date1++;
	}
	
	// fill poster
	robot_ptr_t robotPtr = worldPtr->mapList().front()->robotList().front();
	double time = kernel::Clock::getTime();
	jblas::vec x(6);
	jblas::sym_mat P(6,6);
	robotPtr->getCurrentPose(time-SDI_F->exportDelay, x, P);

	SDI_F->datePomPos = time;
	SDI_F->pomPos.main.euler.yaw = x(5);
	SDI_F->pomPos.main.euler.pitch = x(4);
	SDI_F->pomPos.main.euler.roll = x(3);
	SDI_F->pomPos.main.euler.x = x(0);
	SDI_F->pomPos.main.euler.y = x(1);
	SDI_F->pomPos.main.euler.z = x(2);
	SDI_F->pomPos.main.var.voo = P(5,5);
	SDI_F->pomPos.main.var.vpp = P(4,4);
	SDI_F->pomPos.main.var.vrr = P(3,3);
	SDI_F->pomPos.main.var.vxx = P(0,0);
	SDI_F->pomPos.main.var.vyy = P(1,1);
	SDI_F->pomPos.main.var.vzz = P(2,2);
	
	SDI_F->pomPos.main.var.vop = P(5,4);
	SDI_F->pomPos.main.var.vor = P(5,3);
	SDI_F->pomPos.main.var.vpr = P(4,3);
	SDI_F->pomPos.main.var.vox = P(5,0);
	SDI_F->pomPos.main.var.vpx = P(4,0);
	SDI_F->pomPos.main.var.vrx = P(3,0);
	SDI_F->pomPos.main.var.voy = P(5,1);
	SDI_F->pomPos.main.var.vpy = P(4,1);
	SDI_F->pomPos.main.var.vry = P(3,1);
	SDI_F->pomPos.main.var.vxy = P(0,1);
	SDI_F->pomPos.main.var.voz = P(5,2);
	SDI_F->pomPos.main.var.vpz = P(4,2);
	SDI_F->pomPos.main.var.vrz = P(3,2);
	SDI_F->pomPos.main.var.vxz = P(0,2);
	SDI_F->pomPos.main.var.vyz = P(1,2);
	
	return OK;
}


/*------------------------------------------------------------------------
 * Init
 *
 * Description: 
 *
 * Reports:      OK
 *              S_rtslam_FWBUS
 *              S_rtslam_CONFIGURED
 *              S_rtslam_NO_CAMERA
 *              S_rtslam_UNKNOWN_ERROR
 */

/* rtslamInitMain  -  codel EXEC of Init
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
rtslamInitMain(RtslamInitParams *init_params, int *report)
{

	/// init options
	
	intOpts[iDispQt] = ((int)init_params->display) & RTSLAM_DISPLAY_2D;
	intOpts[iDispGdhe] = ((int)init_params->display) & RTSLAM_DISPLAY_3D;
	intOpts[iRenderAll] = 0;
	intOpts[iReplay] = 0;
	intOpts[iDump] = (int)init_params->dump;
	intOpts[iRandSeed] = 0;
	intOpts[iPause] = 0;
	intOpts[iVerbose] = 5;
	intOpts[iMap] = (int)init_params->map;
	intOpts[iRobot] = (int)init_params->robot;
	intOpts[iCamera] = (int)init_params->camera;
	intOpts[iTrigger] = (int)init_params->trigger;
	intOpts[iGps] = (int)init_params->gps;
	intOpts[iOdom] = (int)init_params->odo;
	intOpts[iExtloc] = (int)init_params->extloc;
	intOpts[iSimu] = 0;
	intOpts[iExport] = 0;
	intOpts[iCamsFilter] = 100000;
	
	floatOpts[fFreq] = (double)init_params->freq;
	floatOpts[fShutter] = (double)init_params->shutter;
	floatOpts[fHeading] = 0.0;
	
	strOpts[sDataPath] = init_params->data_path;
	strOpts[sConfigSetup] = init_params->config_setup;
	strOpts[sConfigEstimation] = init_params->config_estimation;
	strOpts[sLog] = init_params->log_file;

	// pom frame poster
	if (posterFind(pomPoster, &pomFramePosterId) == ERROR)
	{
		printf("[rtslam-genom] Cannot read pom tagging poster %s, not using it.\n", pomPoster);
		pomFramePosterId = NULL;
	}
	
	/// build objects
	try {
		if (!demo_slam_init())
		{
			switch (worldPtr->error) {
				case eParam: *report = S_rtslam_PARAM;
				case eConfig: *report = S_rtslam_CONFIG;
				case eDependency: *report = S_rtslam_DEPENDENCY;
				case eNotSupported: *report = S_rtslam_NOT_SUPPORTED;
			}
			SDI_F->status.ready = RTSLAM_FALSE;
			SDI_F->status.error_code = worldPtr->error;
			return ETHER;
		}
	} catch (jafar::kernel::Exception &e) { *report = S_rtslam_EXCEPTION; return ETHER; }

	return ETHER;
}

/*------------------------------------------------------------------------
 * Start
 *
 * Description: 
 *
 * Reports:      OK
 */

static boost::thread *thread_slam = NULL;

/* rtslamStartStart  -  codel START of Start
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
rtslamStartStart(int *report)
{
	//GENOM_UNLOCK_SDI; demo_slam_run();
	thread_slam = new boost::thread(demo_slam_run);
	return EXEC;
}


/* rtslamStartMain  -  codel EXEC of Start
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
rtslamStartMain(int *report)
{
	SDI_F->status.ready = (ready ? RTSLAM_TRUE : RTSLAM_FALSE);
	
	if (worldPtr->error != eNoError) {
		switch (worldPtr->error) {
			case eNoSensorData: *report = S_rtslam_NO_DATA;
			case eCrashed: *report = S_rtslam_CRASHED;
		}
		SDI_F->status.ready = RTSLAM_FALSE;
		SDI_F->status.error_code = worldPtr->error;
		return INTER;
	}
	if (thread_slam->timed_join(boost::posix_time::milliseconds(0)))
	{
		*report = S_rtslam_EXCEPTION;
		SDI_F->status.ready = RTSLAM_FALSE;
		SDI_F->status.error_code = worldPtr->error;
		return INTER;
	}
	if (SDI_F->pomPos.main.var.vxx+SDI_F->pomPos.main.var.vyy+SDI_F->pomPos.main.var.vzz >
		LARGE_UNCERTAINTY_THRESHOLD*LARGE_UNCERTAINTY_THRESHOLD)
	{
		*report = S_rtslam_LARGE_UNCERTAINTY;
		SDI_F->status.ready = RTSLAM_FALSE; // maybe should not do that to keep exporting
		SDI_F->status.error_code = worldPtr->error = eLargeUncertainty;
		return ETHER;
	}
	return EXEC;
}


/* rtslamStartInter  -  codel INTER of Start
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
rtslamStartInter(int *report)
{
	worldPtr->exit(true);
	SDI_F->status.ready = RTSLAM_FALSE;
	thread_slam->join();
	delete thread_slam;
	return ETHER;
}



/*------------------------------------------------------------------------
 * Log
 *
 * Description:
 *
 * Reports:      OK
 *              S_rtslam_CANNOT_OPEN_FILE
 *              S_rtslam_CANNOT_WRITE_FILE
 */

static FILE *rtslamLogFile = NULL;

/* rtslamLogStart  -  codel START of Log
	Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
rtslamLogStart(GEN_STRING128 *fileName, int *report)
{
	if (!(rtslamLogFile = fopen(fileName->name, "a")))
	{
		printf ("rtslamLogStart: cannot open log file %s\n", fileName->name);
		*report = S_rtslam_CANNOT_OPEN_FILE;
		return ETHER;
	}

	if (!fprintf(rtslamLogFile, "#date\tdate1\tdate2\tx\ty\tz\tyaw\tpitch\troll\txStd\tyStd\tzStd\tyawStd\tpitchStd\trollStd\txyCov\n"))
	{
		printf ("rtslamLogStart: cannot log in file %s\n", fileName->name);
		*report = S_rtslam_CANNOT_WRITE_FILE;
		return INTER; // calls rtslamLogInter
	}

	printf ("rtslamLogStart: log file %s found\n", fileName->name);
	return EXEC;
}

/* rtslamLogExec  -  codel EXEC of Log
	Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
rtslamLogExec(GEN_STRING128 *fileName, int *report)
{
	if (!fprintf(rtslamLogFile, "%lf\t%d\t%d\t%.6f\t%.6f\t%.6f\t%.6g\t%.6g\t%.6g\t%.6g\t%.6g\t%.6g\t%.6g\t%.6g\t%.6g\t%.6g\n",
		SDI_F->datePomPos, SDI_F->pomPos.date1, SDI_F->pomPos.date2,
		SDI_F->pomPos.main.euler.x, SDI_F->pomPos.main.euler.y, SDI_F->pomPos.main.euler.z,
		SDI_F->pomPos.main.euler.yaw, SDI_F->pomPos.main.euler.pitch, SDI_F->pomPos.main.euler.roll,
		sqrt(SDI_F->pomPos.main.var.vxx), sqrt(SDI_F->pomPos.main.var.vyy), sqrt(SDI_F->pomPos.main.var.vzz),
		sqrt(SDI_F->pomPos.main.var.voo), sqrt(SDI_F->pomPos.main.var.vpp), sqrt(SDI_F->pomPos.main.var.vrr),
		SDI_F->pomPos.main.var.vxy
		))
	{
		printf ("rtslamLogExec: cannot log in file %s\n", fileName->name);
		*report = S_rtslam_CANNOT_WRITE_FILE;
		return INTER; // calls rtslamLogInter
	}

	return EXEC;
}


/* rtslamLogInter  -  codel INTER of Log
	Returns:  INTER ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
rtslamLogInter(GEN_STRING128 *fileName, int *report)
{
	if (rtslamLogFile)
	{
		fclose(rtslamLogFile);
		rtslamLogFile = NULL;
	}
	return ETHER;
}



/*------------------------------------------------------------------------
 * SetInitialPose
 *
 * Description: 
 *
 * Reports:      OK
 */
ACTIVITY_EVENT
rtslamSetInitialPoseMain(RtslamPoseParams *pose, int *report)
{
	map_ptr_t mapPtr = worldPtr->mapList().front();
	robot_ptr_t robPtr = mapPtr->robotList().front();
	robPtr->setInitialPose(
		pose->pos.x,pose->pos.y,pose->pos.z,pose->ori.roll,pose->ori.pitch,pose->ori.yaw,
		pose->pos.xStd,pose->pos.yStd,pose->pos.zStd,pose->ori.rollStd,pose->ori.pitchStd,pose->ori.yawStd);

//configSetup.UNCERT_ATTITUDE,configSetup.UNCERT_HEADING
	
	return ETHER;
}


/*------------------------------------------------------------------------
 * SetOrigin
 *
 * Description:
 *
 * Reports:      OK
 */
ACTIVITY_EVENT
rtslamSetOriginMain(RtslamPosParams *origin, int *report)
{
	map_ptr_t mapPtr = worldPtr->mapList().front();
	robot_ptr_t robPtr = mapPtr->robotList().front();
	robPtr->setOrigin(origin->x, origin->y, origin->z);

	return ETHER;
}


/*------------------------------------------------------------------------
 * SetSensorState
 *
 * Description:
 *
 * Reports:      OK
 */
ACTIVITY_EVENT
rtslamSetSensorState(RtslamSensorStateParams *state, int *report)
{
	map_ptr_t mapPtr = worldPtr->mapList().front();
	robot_ptr_t robPtr = mapPtr->robotList().front();
	bool found = false;
	
	for (RobotAbstract::SensorList::iterator senIter = robPtr->sensorList().begin();
		senIter != robPtr->sensorList().end(); ++senIter)
	{
		if (strcmp((*senIter)->name().c_str(), state->name.name) == 0)
		{
			found = true;
			switch (state->state) {
				case RTSLAM_STATE_ON:
					(*senIter)->enable(true);
					(*senIter)->enableHard(true);
					break;
				case RTSLAM_STATE_OFF:
					(*senIter)->enable(false);
					(*senIter)->enableHard(false);
					break;
				case RTSLAM_STATE_FAIL:
					(*senIter)->enable(true);
					(*senIter)->enableHard(false);
					break;
			}
		}
	}

	if (!found) *report = S_rtslam_SENSOR_NOT_FOUND;
	return ETHER;
}


/*------------------------------------------------------------------------
 * rtslamReadExternalLoc
 *
 * Description:
 *
 * Reports:      OK
 */
ACTIVITY_EVENT
rtslamReadExternalLoc(RtslamReadExternalLocParams *extloc, int *report)
{
	map_ptr_t mapPtr = worldPtr->mapList().front();
	robot_ptr_t robPtr = mapPtr->robotList().front();
	bool found = false;

	for (RobotAbstract::SensorList::iterator senIter = robPtr->sensorList().begin();
		senIter != robPtr->sensorList().end(); ++senIter)
	{
		if (strcmp((*senIter)->name().c_str(), "extloc") == 0)
		{
			found = true;
			hardware::HardwareSensorExternalLoc *hard = static_cast<hardware::HardwareSensorExternalLoc*>(static_cast<SensorProprioAbstract*>((*senIter).get())->hardwareSensorPtr.get());
			hard->loadACTION(extloc->config_file.name, extloc->state_file.name, extloc->detect_file.name, extloc->static_file.name);
		}
	}

	if (!found) *report = S_rtslam_SENSOR_NOT_FOUND;
	return ETHER;
}


/*------------------------------------------------------------------------
 * rtslamStopDumping
 *
 * Description:
 *
 * Reports:      OK
 */
ACTIVITY_EVENT
rtslamStopDumping(int *report)
{
	map_ptr_t mapPtr = worldPtr->mapList().front();
	robot_ptr_t robPtr = mapPtr->robotList().front();
	robPtr->stopDumping();
	return ETHER;
}



/*------------------------------------------------------------------------
 * SetInitialOrientation
 *
 * Description:
 *
 * Reports:      OK
 */
ACTIVITY_EVENT
rtslamSetInitialOrientationMain(RtslamOrientationParams *ori, int *report)
{
	map_ptr_t mapPtr = worldPtr->mapList().front();
	robot_ptr_t robPtr = mapPtr->robotList().front();
	robPtr->setInitialOrientation(
		ori->roll,ori->pitch,ori->yaw,ori->rollStd,ori->pitchStd,ori->yawStd);

//configSetup.UNCERT_ATTITUDE,configSetup.UNCERT_HEADING

	return ETHER;
}

