
#ifndef RTSLAM_STRUCT_H_
#define RTSLAM_STRUCT_H_

//#include "rtslam/worldAbstract.hpp"

#define RTSLAM_OBJECT_LENGTH 32
#define RTSLAM_DEVICE_LENGTH 64
#define RTSLAM_FILE_LENGTH   256
#define RTSLAM_POSTER_LENGTH 64

/*
typedef char RTSLAM_OBJECT_NAME[RTSLAM_OBJECT_LENGTH];
typedef char RTSLAM_DEVICE_NAME[RTSLAM_DEVICE_LENGTH];
typedef char RTSLAM_FILE_NAME  [RTSLAM_FILE_LENGTH  ];
typedef char RTSLAM_POSTER_NAME[RTSLAM_POSTER_LENGTH];
*/


// PARAMETERS

typedef enum RtslamBoolean {
	RTSLAM_FALSE = 0, RTSLAM_TRUE = 1
} RtslamBoolean;

typedef enum RtslamEnable {
	RTSLAM_OFF = 0, RTSLAM_ON = 1
} RtslamEnable;

typedef enum RtslamMapType {
	RTSLAM_MAP_ODOMETRY = 0, RTSLAM_MAP_GLOBAL = 1, RTSLAM_MAP_MULTIMAP = 2
} RtslamMapType;

typedef enum RtslamRobotType {
	RTSLAM_ROBOT_CONSTVEL = 0, RTSLAM_ROBOT_INERTIAL = 1
} RtslamRobotType;

typedef enum RtslamCameraType {
	RTSLAM_CAM_NONE = 0, RTSLAM_CAM_LEFT_RAW = 10, RTSLAM_CAM_RIGHT_RAW = 20, RTSLAM_CAM_BI_RAW = 30
} RtslamCameraType;

typedef enum RtslamTriggerType {
	RTSLAM_TRIGGER_INTERNAL = 0, RTSLAM_TRIGGER_EXTERNAL_M1 = 1, RTSLAM_TRIGGER_EXTERNAL_M0 = 2, RTSLAM_TRIGGER_EXTERNAL_M14 = 3
} RtslamTriggerType;

typedef enum RtslamDisplayType {
	RTSLAM_DISPLAY_OFF = 0, RTSLAM_DISPLAY_2D = 1, RTSLAM_DISPLAY_3D = 2, RTSLAM_DISPLAY_ON = 3
} RtslamDisplayType;

typedef enum RtslamGpsType {
	RTSLAM_GPS_OFF = 0, RTSLAM_GPS_POS = 1, RTSLAM_GPS_POSVEL = 2, RTSLAM_GPS_POSORI = 3
} RtslamGpsType;

typedef enum RtslamOdoType {
	RTSLAM_ODO_OFF = 0, RTSLAM_ODO_ON = 1
} RtslamOdoType;

typedef enum RtslamExtlocType {
	RTSLAM_EXTLOC_DISABLED = 0, RTSLAM_EXTLOC_LIVE_BUNDLE = 10, RTSLAM_EXTLOC_LIVE_FULLPOSE = 11, RTSLAM_EXTLOC_FILE_BUNDLE = 20, RTSLAM_EXTLOC_FILE_FULLPOSE = 21
} RtslamExtlocType;

typedef enum RtslamSensorState {
	RTSLAM_STATE_ON = 0, ///< sensor is enabled
	RTSLAM_STATE_OFF,    ///< sensor is disabled and not expected to work, but it is logged
	RTSLAM_STATE_FAIL    ///< simulates sensor failure: it has no data, but we are expecting some
} RtslamSensorState;

typedef struct RtslamInitParams {
	RtslamMapType map;
	RtslamRobotType robot;
	double freq;
	double shutter;
	RtslamCameraType camera;
	RtslamTriggerType trigger;
	RtslamGpsType gps;
	RtslamOdoType odo;
	RtslamExtlocType extloc;
	RtslamEnable dump;
	RtslamDisplayType display;
	char data_path[RTSLAM_FILE_LENGTH+4];
	char log_file[RTSLAM_FILE_LENGTH];
	char config_setup[RTSLAM_FILE_LENGTH];
	char config_estimation[RTSLAM_FILE_LENGTH];
} RtslamInitParams;

typedef struct RtslamPosParams {
	double x;
	double y;
	double z;
} RtslamPosParams;

typedef struct RtslamPositionParams {
	double x; // x position
	double y; // y position
	double z; // y position
	double xStd;
	double yStd;
	double zStd;
} RtslamPositionParams;

typedef struct RtslamOrientationParams {
	double yaw; // yaw angle
	double pitch; // pitch angle
	double roll; // roll angle
	double yawStd;
	double pitchStd;
	double rollStd;
} RtslamOrientationParams;

typedef struct RtslamPoseParams {
	RtslamPositionParams pos;
	RtslamOrientationParams ori;
} RtslamPoseParams;


typedef struct RtslamSensorStateParams {
	GEN_STRING128 name;
	RtslamSensorState state;
} RtslamSensorStateParams;

typedef struct RtslamReadExternalLocParams {
	GEN_STRING128 config_file;
	GEN_STRING128 state_file;
	GEN_STRING128 detect_file;
	GEN_STRING128 static_file;
} RtslamReadExternalLocParams;


// POSTERS

typedef struct RtslamStatus
{
	RtslamBoolean ready;
	double time_to_ready;
	int error_code; ///< see definition in WorldAbstract::ErrorType
} RtslamStatus;



#endif
